@@ -8,11 +8,10 @@
   +--------------------------------------------------------------------+
   | Copyright (c) 2006, Thilo Raufeisen <traufeisen@php.net>           |
   | Copyright (c) 2013, Jim Jagielski <jimjag@php.net>                 |
+  | Copyright (c) 2016, Jakub Zelenka <bukka@php.net>                  |
   +--------------------------------------------------------------------+
 */
 
-/* $Id: php_gnupg_keylistiterator.h 330956 2013-07-17 13:00:21Z jimjag $ */
-
 #ifndef PHP_GNUPG_KEYLISTITERATOR_H
 #define PHP_GNUPG_KEYLISTITERATOR_H
 
@@ -29,19 +28,17 @@ extern zend_module_entry gnupg_keyiterator_module_entry;
 #endif
 
 #include <gpgme.h>
+#include "phpc/phpc.h"
 
 #define gnupg_keylistiterator_init() _gnupg_keylistiterator_init(INIT_FUNC_ARGS_PASSTHRU)
-extern int  _gnupg_keylistiterator_init(INIT_FUNC_ARGS);
+extern int _gnupg_keylistiterator_init(INIT_FUNC_ARGS);
 
-typedef struct _gnupg_keylistiterator_object{
-	zend_object zo;
+PHPC_OBJ_STRUCT_BEGIN(gnupg_keylistiterator)
 	gpgme_ctx_t ctx;
-    gpgme_error_t err;
-    gpgme_key_t gpgkey;
-    zval pattern;
-} gnupg_keylistiterator_object;
-
-static zend_class_entry *gnupg_keylistiterator_class_entry;
+	gpgme_error_t err;
+	gpgme_key_t gpgkey;
+	char *pattern;
+PHPC_OBJ_STRUCT_END()
 
 PHP_METHOD(gnupg_keylistiterator, __construct);
 PHP_METHOD(gnupg_keylistiterator, current);
@@ -50,7 +47,6 @@ PHP_METHOD(gnupg_keylistiterator, rewind);
 PHP_METHOD(gnupg_keylistiterator, key);
 PHP_METHOD(gnupg_keylistiterator, valid);
 
-
 #ifdef ZTS
 #define GNUPG_G(v) TSRMG(gnupg_globals_id, zend_gnupg_globals *, v)
 #else
