@@ -50,33 +50,19 @@
 #define PHPC_OBJ_INIT_HANDLERS(_name) \
 	memcpy(&PHPC_OBJ_GET_HANDLER_VAR_NAME(_name), \
 		zend_get_std_object_handlers(), sizeof(zend_object_handlers))
-#define PHPC_CLASS_SET_HANDLER_CREATE(_class_entry, _name) \
-	_class_entry.create_object = PHPC_OBJ_GET_HANDLER_FCE(_name, create)
-#define PHPC_OBJ_SET_HANDLER_CLONE(_name) \
-	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).clone_obj = PHPC_OBJ_GET_HANDLER_FCE(_name, clone)
-#define PHPC_OBJ_SET_HANDLER_COMPARE(_name) \
-	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).compare_objects = PHPC_OBJ_GET_HANDLER_FCE(_name, compare)
-#define PHPC_OBJ_HAS_HANDLER_GET_GC (PHP_VERSION_ID > 50399)
-#if PHPC_OBJ_HAS_HANDLER_GET_GC
-#define PHPC_OBJ_SET_HANDLER_GET_GC(_name) \
-	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).get_gc = PHPC_OBJ_GET_HANDLER_FCE(_name, get_gc)
-#else
-#define PHPC_OBJ_SET_HANDLER_GET_GC(_name) PHPC_NOOP
+
+/* for arginfo */
+#ifndef ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX
+#define ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(name, return_reference, required_num_args, type, allow_null) \
+        ZEND_BEGIN_ARG_INFO_EX(name, 0, return_reference, required_num_args)
 #endif
-#define PHPC_OBJ_HAS_HANDLER_GET_DEBUG_INFO (PHP_VERSION_ID > 50299)
-#if PHPC_OBJ_HAS_HANDLER_GET_DEBUG_INFO
-#define PHPC_OBJ_SET_HANDLER_GET_DEBUG_INFO(_name) \
-	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).get_debug_info = PHPC_OBJ_GET_HANDLER_FCE(_name, get_debug_info)
-#else
-#define PHPC_OBJ_SET_HANDLER_GET_DEBUG_INFO(_name) PHPC_NOOP
+
+/* ZEND_ACC_CTOR and ZEND_ACC_DTOR is removed in 7.4 */
+#ifndef ZEND_ACC_CTOR
+#define ZEND_ACC_CTOR 0
 #endif
-/* there is such handler in 5.2 but we would have to re-implement zend_std_get_properties */
-#define PHPC_OBJ_HAS_HANDLER_GET_PROPERTIES (PHP_VERSION_ID > 50299)
-#if PHPC_OBJ_HAS_HANDLER_GET_PROPERTIES
-#define PHPC_OBJ_SET_HANDLER_GET_PROPERTIES(_name) \
-	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).get_properties = PHPC_OBJ_GET_HANDLER_FCE(_name, get_properties)
-#else
-#define PHPC_OBJ_SET_HANDLER_GET_PROPERTIES(_name) PHPC_NOOP
+#ifndef ZEND_ACC_DTOR
+#define ZEND_ACC_DTOR 0
 #endif
 
 #if PHP_VERSION_ID < 50299
@@ -85,20 +71,25 @@
 #define phpc_function_entry const zend_function_entry
 #endif
 
-
-/* initializing properties in obejct (object_properties_init was added in PHP 5.4) */
 #if PHP_VERSION_ID < 50399
-#define PHPC_OBJ_PROPERTIES_INIT(zo, class_type) \
-	{ \
-		zval *tmp; \
-		zend_hash_copy((*(zo)).properties, \
-			&(class_type)->default_properties, \
+/* initializing properties in obejct (object_properties_init was added in PHP 5.4) */
+#define PHPC_OBJ_PROPERTIES_INIT(_zo, _class_type) \
+	do { \
+		zval *_phpc_tmp; \
+		zend_hash_copy((*(_zo)).properties, \
+			&(_class_type)->default_properties, \
 			(copy_ctor_func_t) zval_add_ref, \
-			(void *) &tmp, \
+			(void *) &_phpc_tmp, \
 			sizeof(zval *)); \
-	}
+	} while(0)
+
+/* write property */
+#define PHPC_OBJ_STD_WRITE_PROPERTY(_obj, _member, _value) \
+	Z_OBJ_HT_P(_obj)->write_property(_obj, _member, _value TSRMLS_CC)
 #else
 #define PHPC_OBJ_PROPERTIES_INIT object_properties_init
+#define PHPC_OBJ_STD_WRITE_PROPERTY(_obj, _member, _value) \
+	zend_std_write_property(_obj, _member, _value, NULL TSRMLS_CC)
 #endif
 
 /* common fcall macros */
@@ -127,6 +118,47 @@
 	PHPC_LONG_TO_INT_EX2(_plv, _lv, _lv = INT_MAX, _lv = INT_MIN)
 
 
+#if PHP_MAJOR_VERSION < 8
+/* PHP 5 and 7 */
+/* object */
+#define phpc_obj_t zval
+#define PHPC_OBJ_TO_POBJ(_obj) Z_OBJ_P(_obj)
+#define PHPC_OBJ_FOR_PROP(_obj) (_obj)
+#define PHPC_OBJ_HANDLER_COMPARE_NAME compare_objects
+
+#define PHPC_FCALL_FCI_INIT(_fci, _callback, _count, _no_separ) \
+	do { \
+		(_fci).params = PHPC_FCALL_PARAMS_NAME(_callback); \
+		(_fci).param_count = (_count); \
+		(_fci).no_separation = (_no_separ); \
+	} while (0)
+#else
+/* PHP 8 */
+
+/* object */
+#define phpc_obj_t zend_object
+#define PHPC_OBJ_TO_POBJ(_obj) (_obj)
+#define PHPC_OBJ_FOR_PROP(_obj) Z_OBJ_P(_obj)
+#define PHPC_OBJ_HANDLER_COMPARE_NAME compare
+
+/* fcall */
+#define PHPC_FCALL_FCI_INIT(_fci, _callback, _count, _no_separ) \
+	do { \
+		(_fci).params = PHPC_FCALL_PARAMS_NAME(_callback); \
+		(_fci).param_count = (_count); \
+	} while (0)
+/* ZTS */
+#define TSRMLS_D void
+#define TSRMLS_DC
+#define TSRMLS_C
+#define TSRMLS_CC
+#define TSRMLS_FETCH()
+#endif
+
+#define PHPC_OBJ_STD_GET_PROPERTIES(_obj) \
+	zend_std_get_properties((_obj) TSRMLS_CC)
+
+
 #if PHP_MAJOR_VERSION == 5
 
 /* MODULE */
@@ -195,6 +227,12 @@ typedef int phpc_str_size_t;
 #define PHPC_STR_FROM_PTR_VAL(_str, _strpv) \
 	PHPC_STR_VAL(_str) = *PHPC_STR_VAL(_strpv); \
 	PHPC_STR_LEN(_str) = strlen(*PHPC_STR_VAL(_strpv))
+#define PHPC_STR_FROM_ZVAL(_str, _zv) \
+	PHPC_STR_VAL(_str) = Z_STRVAL(_zv); \
+	PHPC_STR_LEN(_str) = Z_STRLEN(_zv)
+#define PHPC_STR_FROM_PZVAL(_str, _pzv) \
+	PHPC_STR_VAL(_str) = Z_STRVAL_P(_pzv); \
+	PHPC_STR_LEN(_str) = Z_STRLEN_P(_pzv)
 #define PHPC_STR_RETURN(_name) \
 	RETURN_STRINGL(PHPC_STR_VAL(_name), PHPC_STR_LEN(_name), 0)
 
@@ -221,6 +259,9 @@ typedef int phpc_str_size_t;
 #define PHPC_CSTRL_RETVAL(_name, _len) RETVAL_STRINGL(_name, _len, 1)
 #define PHPC_CSTR_RETVAL(_name)        RETVAL_STRING(_name, 1)
 
+/* ZSTR */
+#define PHPC_ZSTR_VAL(_name) _name
+
 /* Smart string */
 #if defined(PHPC_SMART_STR_INCLUDE) || defined(PHPC_SMART_CSTR_INCLUDE)
 #include "ext/standard/php_smart_str.h"
@@ -254,6 +295,7 @@ typedef long phpc_res_value_t;
 	zend_fetch_resource(&(_pz_res) TSRMLS_CC, -1, _res_type_name, NULL, 2, _res_type_1, _res_type_2)
 #define PHPC_RES_DELETE(_pz_res) \
 	zend_list_delete(Z_LVAL_P(_pz_res))
+#define PHPC_RES_CLOSE PHPC_RES_DELETE
 
 /* resource to zval */
 #define PHPC_RES_PZVAL(_res, _pzv) \
@@ -281,10 +323,7 @@ typedef long phpc_res_value_t;
 	(PHPC_OBJ_STRUCT_NAME(_name) *) _object
 #define PHPC_OBJ_FROM_ZVAL(_name, _zv) \
 	(PHPC_OBJ_STRUCT_NAME(_name) *) zend_object_store_get_object(_zv TSRMLS_CC)
-#define PHPC_OBJ_FROM_SELF(_name) \
-	PHPC_OBJ_FROM_ZVAL(_name, PHPC_SELF)
-#define PHP_OBJ_GET_HANDLER_OBJ_FROM_ZOBJ(_name) \
-	PHPC_OBJ_FROM_ZOBJ(_name, _phpc_object)
+#define PHPC_OBJ_FROM_POBJ PHPC_OBJ_FROM_ZVAL
 
 /* create_ex object handler helper */
 #define PHPC_OBJ_HANDLER_CREATE_EX(_name) \
@@ -321,7 +360,7 @@ typedef long phpc_res_value_t;
 
 /* clone object handler */
 #define PHPC_OBJ_HANDLER_CLONE(_name) \
-	PHPC_OBJ_DEFINE_HANDLER_FCE(zend_object_value, _name, clone)(zval *PHPC_SELF TSRMLS_DC)
+	PHPC_OBJ_DEFINE_HANDLER_FCE(zend_object_value, _name, clone)(phpc_obj_t *PHPC_SELF TSRMLS_DC)
 #define PHPC_OBJ_HANDLER_CLONE_DECLARE() zend_object_value _phpc_retval
 #define PHPC_OBJ_HANDLER_CLONE_MEMBERS(_name, _new_obj, _old_obj) \
 	do { \
@@ -346,8 +385,8 @@ typedef long phpc_res_value_t;
 	PHPC_OBJ_STRUCT_DECLARE(_name, _obj) = PHPC_OBJ_FROM_ZVAL(_name, _phpc_obj ## _id)
 
 /* handler setters */
-#define PHPC_OBJ_SET_HANDLER_OFFSET(_name) PHPC_NOOP
-#define PHPC_OBJ_SET_HANDLER_FREE(_name) PHPC_NOOP
+#define PHPC_OBJ_SET_SPECIFIC_HANDLER_OFFSET(_handlers, _name) PHPC_NOOP
+#define PHPC_OBJ_SET_SPECIFIC_HANDLER_FREE(_handlers, _name) PHPC_NOOP
 
 /* read propery */
 #define PHPC_READ_PROPERTY_RV_NAME
@@ -358,6 +397,7 @@ typedef long phpc_res_value_t;
 /* HASH */
 
 /* apply count */
+#define PHPC_HASH_HAS_APPLY_COUNT(_ht)  (_ht->nApplyCount > 0)
 #define PHPC_HASH_GET_APPLY_COUNT(_ht)  _ht->nApplyCount
 #define PHPC_HASH_INC_APPLY_COUNT(_ht)  _ht->nApplyCount++
 #define PHPC_HASH_DEC_APPLY_COUNT(_ht)  _ht->nApplyCount--
@@ -382,6 +422,12 @@ typedef long phpc_res_value_t;
 #define PHPC_HASH_CSTR_ADD(_ht, _cstr_value, _pzv) \
 	zend_hash_add(_ht, _cstr_value, strlen(_cstr_value) + 1, &_pzv, sizeof(_pzv), NULL)
 
+/* next insert */
+#define PHPC_HASH_NEXT_INDEX_INSERT_PTR(_ht, _ptr, _ptr_size) \
+	zend_hash_next_index_insert(_ht, _ptr, _ptr_size, NULL)
+#define PHPC_HASH_NEXT_INDEX_INSERT(_ht, _pzv) \
+	zend_hash_next_index_insert(_ht, _pzv, sizeof(_pzv), NULL)
+
 /* update ptr */
 #define PHPC_HASH_INDEX_UPDATE_PTR(_ht, _idx, _ptr, _ptr_size) \
 	zend_hash_index_update(_ht, _idx, _ptr, _ptr_size, NULL)
@@ -458,6 +504,13 @@ typedef long phpc_res_value_t;
 #define PHPC_HASH_COPY_EX(_target, _source, _copy_ctr) \
 	zend_hash_copy(_target, _source, NULL, NULL, sizeof(zval *))
 
+/* key result constant has been renamed in 5.4 */
+#if PHP_VERSION_ID < 50399
+#define _PHPC_HASH_KEY_NON_EXISTENT HASH_KEY_NON_EXISTANT
+#else
+#define _PHPC_HASH_KEY_NON_EXISTENT HASH_KEY_NON_EXISTENT
+#endif
+
 /* iteration for each element */
 #define PHPC_HASH_FOREACH_KEY(_ht, _h, _key) do { \
 	HashPosition _pos; \
@@ -465,7 +518,7 @@ typedef long phpc_res_value_t;
 	int _key_type; \
 	for (zend_hash_internal_pointer_reset_ex((_ht), &_pos); \
 			(_key_type = zend_hash_get_current_key_ex( \
-				(_ht), &PHPC_STR_VAL(_key), &_str_length, &_h, 0, &_pos)) != HASH_KEY_NON_EXISTENT; \
+				(_ht), &PHPC_STR_VAL(_key), &_str_length, &_h, 0, &_pos)) != _PHPC_HASH_KEY_NON_EXISTENT; \
 			zend_hash_move_forward_ex((_ht), &_pos) ) { \
 		if (_key_type == HASH_KEY_IS_STRING) { \
 			PHPC_STR_LEN(_key) = (int) _str_length - 1; \
@@ -528,7 +581,15 @@ typedef long phpc_res_value_t;
 	} while(0)
 
 
+/* SYMTABLE */
+
+/* update*/
+#define PHPC_SYMTABLE_UPDATE(_ht, _str, _pzv) \
+	zend_symtable_update(_ht, PHPC_STR_VAL(_str), PHPC_STR_LEN(_str) + 1, &_pzv, sizeof(_pzv), NULL)
+
+
 /* ARRAY */
+
 #define PHPC_ARRAY_ADD_ASSOC_NULL_EX(_arr, _key, _key_len) \
     add_assoc_null_ex(_arr, _key, (_key_len) + 1)
 #define PHPC_ARRAY_ADD_ASSOC_BOOL_EX(_arr, _key, _key_len, _b) \
@@ -605,29 +666,44 @@ add_assoc_resource_ex(_arr, _key, (_key_len) + 1, _r)
 /* ZVAL */
 typedef zval * phpc_val;
 
-#define PHPC_TYPE        Z_TYPE_P
-#define PHPC_TYPE_P      Z_TYPE_PP
+#define PHPC_TYPE           Z_TYPE_P
+#define PHPC_TYPE_P         Z_TYPE_PP
 
-#define PHPC_LVAL        Z_LVAL_P
-#define PHPC_LVAL_P      Z_LVAL_PP
+#define PHPC_LVAL           Z_LVAL_P
+#define PHPC_LVAL_P         Z_LVAL_PP
 
-#define PHPC_DVAL        Z_DVAL_P
-#define PHPC_DVAL_P      Z_DVAL_PP
+#define PHPC_DVAL           Z_DVAL_P
+#define PHPC_DVAL_P         Z_DVAL_PP
 
-#define PHPC_STRVAL      Z_STRVAL_P
-#define PHPC_STRVAL_P    Z_STRVAL_PP
+#define PHPC_STRVAL         Z_STRVAL_P
+#define PHPC_STRVAL_P       Z_STRVAL_PP
 
-#define PHPC_STRLEN      Z_STRLEN_P
-#define PHPC_STRLEN_P    Z_STRLEN_PP
+#define PHPC_STRLEN         Z_STRLEN_P
+#define PHPC_STRLEN_P       Z_STRLEN_PP
 
-#define PHPC_ARRVAL      Z_ARRVAL_P
-#define PHPC_ARRVAL_P    Z_ARRVAL_PP
+#define PHPC_ARRVAL         Z_ARRVAL_P
+#define PHPC_ARRVAL_P       Z_ARRVAL_PP
 
-#define PHPC_OBJCE       Z_OBJCE_P
-#define PHPC_OBJCE_P     Z_OBJCE_PP
+#define PHPC_OBJ_HT         Z_OBJ_HT_P
+#define PHPC_OBJ_HT_P       Z_OBJ_HT_PP
 
-#define PHPC_OBJPROP     Z_OBJPROP_P
-#define PHPC_OBJPROP_P   Z_OBJPROP_PP
+#define PHPC_OBJ_HANDLER    Z_OBJ_HANDLER_P
+#define PHPC_OBJ_HANDLER_P  Z_OBJ_HANDLER_PP
+
+#define PHPC_OBJ_HANDLE     Z_OBJ_HANDLE_P
+#define PHPC_OBJ_HANDLE_P   Z_OBJ_HANDLE_PP
+
+#define PHPC_OBJCE          Z_OBJCE_P
+#define PHPC_OBJCE_P        Z_OBJCE_PP
+
+#define PHPC_OBJPROP        Z_OBJPROP_P
+#define PHPC_OBJPROP_P      Z_OBJPROP_PP
+
+#define PHPC_OBJDEBUG       Z_OBJDEBUG_P
+#define PHPC_OBJDEBUG_P     Z_OBJDEBUG_PP
+
+#define PHPC_REFCOUNTED(_zv)    1
+#define PHPC_REFCOUNTED_P(_pzv) 1
 
 #define PHPC_PZVAL_CAST_TO_PVAL(_pzv) &_pzv
 
@@ -646,19 +722,39 @@ typedef zval * phpc_val;
 #define PHPC_VAL_COPY(_pv, _zv)   ZVAL_ZVAL(_pv, _zv, 1, 0)
 #define PHPC_VAL_ASSIGN(_pv, _zv) _pv = _zv
 
+#define PHPC_PZVAL_MAKE           MAKE_STD_ZVAL
+#define PHPC_PZVAL_FREE           FREE_ZVAL
+#define PHPC_PZVAL_SET(_pv, _zv)  ZVAL_ZVAL(_pv, _zv, 0, 0)
 #define PHPC_PVAL_DEREF(_pv)      PHPC_NOOP
 #define PHPC_PZVAL_DEREF(_pv)     PHPC_NOOP
 
-#define PHPC_VAL_STR(_pv, _str) \
+#define PHPC_PZVAL_COPY_INIT      INIT_PZVAL_COPY
+#define PHPC_PZVAL_COPY(_pzv_dst, _pzv_src) \
+	*_pzv_dst = *_pzv_src
+
+#define PHPC_VAL_NEW_STR(_pv, _str) \
 	ZVAL_STRINGL(_pv, PHPC_STR_VAL(_str), PHPC_STR_LEN(_str), 0)
+#define PHPC_VAL_STR PHPC_VAL_NEW_STR
 #define PHPC_VAL_CSTR(_pv, _cstr) \
     ZVAL_STRING(_pv, _cstr, 1)
 #define PHPC_VAL_CSTRL(_pv, _cstr, _cstr_len) \
 	ZVAL_STRINGL(_pv, _cstr, _cstr_len, 1)
 
-#define PHPC_PZVAL_STR   PHPC_VAL_STR
-#define PHPC_PZVAL_CSTR  PHPC_VAL_CSTR
-#define PHPC_PZVAL_CSTRL PHPC_VAL_CSTRL
+#define PHPC_PZVAL_NEW_STR PHPC_VAL_NEW_STR
+#define PHPC_PZVAL_STR     PHPC_VAL_STR
+#define PHPC_PZVAL_CSTR    PHPC_VAL_CSTR
+#define PHPC_PZVAL_CSTRL   PHPC_VAL_CSTRL
+
+#define PHPC_ZVAL_COPY(_zv_dst, _zv_src) \
+	_zv_dst = _zv_src
+
+#define PHPC_ZVAL_NEW_STR(_zv, _str) \
+	PHPC_VAL_STR(&_zv, _str)
+#define PHPC_ZVAL_STR PHPC_ZVAL_NEW_STR
+#define PHPC_ZVAL_CSTR(_zv, _cstr) \
+	PHPC_VAL_CSTR(&_zv, _cstr)
+#define PHPC_ZVAL_CSTRL(_zv, _cstr, _cstr_len) \
+	PHPC_VAL_CSTRL(&_zv, _cstr, _cstr_len)
 
 #define PHPC_ZVAL_IS_TRUE(_zv) \
 	(Z_TYPE(_zv) == IS_BOOL && Z_BVAL(_zv))
@@ -802,6 +898,8 @@ typedef size_t    phpc_str_size_t;
 #define PHPC_STR_FROM_PTR_STR(_str, _strp)    _str = *_strp
 #define PHPC_STR_FROM_PTR_VAL(_str, _strpv)   _str = *_strpv
 #define PHPC_STR_RETURN                       RETURN_STR
+#define PHPC_STR_FROM_ZVAL(_str, _zv)         _str = Z_STR(_zv)
+#define PHPC_STR_FROM_PZVAL(_str, _pzv)       _str = Z_STR_P(_pzv)
 /* wrapper macros */
 #define PHPC_STR_INIT(_name, _cstr, _len) \
 	_name = zend_string_init(_cstr, _len, 0)
@@ -818,6 +916,9 @@ typedef size_t    phpc_str_size_t;
 #define PHPC_CSTRL_RETVAL                     RETVAL_STRINGL
 #define PHPC_CSTR_RETVAL                      RETVAL_STRING
 
+/* ZSTR */
+#define PHPC_ZSTR_VAL ZSTR_VAL
+
 /* Smart string */
 #ifdef PHPC_SMART_STR_INCLUDE
 #include "zend_smart_str.h"
@@ -850,6 +951,8 @@ typedef zend_resource * phpc_res_value_t;
 	zend_fetch_resource2(Z_RES_P(_pz_res), _res_type_name, _res_type_1, _res_type_2)
 #define PHPC_RES_DELETE(_pz_res) \
 	zend_list_delete(Z_RES_P(_pz_res))
+#define PHPC_RES_CLOSE(_pz_res) \
+	zend_list_close(Z_RES_P(_pz_res))
 
 /* resource to zval */
 #define PHPC_RES_PZVAL(_res, _pzv) \
@@ -876,10 +979,11 @@ typedef zend_resource * phpc_res_value_t;
 	(PHPC_OBJ_STRUCT_NAME(_name) *)((char*)(_object) - XtOffsetOf(PHPC_OBJ_STRUCT_NAME(_name), std))
 #define PHPC_OBJ_FROM_ZVAL(_name, _zv) \
 	PHPC_OBJ_FROM_ZOBJ(_name, Z_OBJ_P(_zv))
-#define PHPC_OBJ_FROM_SELF(_name) \
-	PHPC_OBJ_FROM_ZVAL(_name, PHPC_SELF)
-#define PHP_OBJ_GET_HANDLER_OBJ_FROM_ZOBJ(_name) \
-	PHPC_OBJ_FROM_ZOBJ(_name, _phpc_object)
+#if PHP_MAJOR_VERSION < 8
+#define PHPC_OBJ_FROM_POBJ PHPC_OBJ_FROM_ZVAL
+#else
+#define PHPC_OBJ_FROM_POBJ PHPC_OBJ_FROM_ZOBJ
+#endif
 
 /* create_ex object handler helper */
 #define PHPC_OBJ_HANDLER_CREATE_EX(_name) \
@@ -911,7 +1015,7 @@ typedef zend_resource * phpc_res_value_t;
 
 /* clone object handler */
 #define PHPC_OBJ_HANDLER_CLONE(_name) \
-	PHPC_OBJ_DEFINE_HANDLER_FCE(zend_object *, _name, clone)(zval *PHPC_SELF)
+	PHPC_OBJ_DEFINE_HANDLER_FCE(zend_object *, _name, clone)(phpc_obj_t *PHPC_SELF)
 #define PHPC_OBJ_HANDLER_CLONE_DECLARE() PHPC_NOOP
 #define PHPC_OBJ_HANDLER_CLONE_MEMBERS(_name, _new_obj, _old_obj) \
 	do { \
@@ -933,26 +1037,34 @@ typedef zend_resource * phpc_res_value_t;
 	PHPC_OBJ_STRUCT_DECLARE(_name, _obj) = PHPC_OBJ_FROM_ZVAL(_name, _phpc_obj ## _id)
 
 /* handler setters */
-#define PHPC_OBJ_SET_HANDLER_OFFSET(_name) \
-	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).offset = XtOffsetOf(PHPC_OBJ_STRUCT_NAME(_name), std)
-#define PHPC_OBJ_SET_HANDLER_FREE(_name) \
-	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).free_obj = PHPC_OBJ_GET_HANDLER_FCE(_name, free)
+#define PHPC_OBJ_SET_SPECIFIC_HANDLER_OFFSET(_handlers, _name) \
+	(_handlers).offset = XtOffsetOf(PHPC_OBJ_STRUCT_NAME(_name), std)
+#define PHPC_OBJ_SET_SPECIFIC_HANDLER_FREE(_handlers, _name) \
+	(_handlers).free_obj = PHPC_OBJ_GET_HANDLER_FCE(_name, free)
 
 /* read propery */
 #define PHPC_READ_PROPERTY_RV_NAME _phpc_read_property_rv
 #define PHPC_READ_PROPERTY_RV_DECLARE zval PHPC_READ_PROPERTY_RV_NAME
 #define PHPC_READ_PROPERTY(_scope, _object, _name, _name_len, _silent) \
-	zend_read_property(_scope, _object, _name, _name_len, _silent, &PHPC_READ_PROPERTY_RV_NAME)
+	zend_read_property(_scope, PHPC_OBJ_FOR_PROP(_object), _name, _name_len, _silent, &PHPC_READ_PROPERTY_RV_NAME)
 
 
 /* HASH */
 
-/* apply count */
-#define PHPC_HASH_GET_APPLY_COUNT  ZEND_HASH_GET_APPLY_COUNT
-#define PHPC_HASH_INC_APPLY_COUNT  ZEND_HASH_INC_APPLY_COUNT
-#define PHPC_HASH_DEC_APPLY_COUNT  ZEND_HASH_DEC_APPLY_COUNT
-#define PHPC_HASH_APPLY_PROTECTION ZEND_HASH_APPLY_PROTECTION
-
+/* recursion protection */
+#if PHP_VERSION_ID < 70299
+#define PHPC_HASH_HAS_APPLY_COUNT(_tht)  (ZEND_HASH_GET_APPLY_COUNT(_tht) > 0)
+#define PHPC_HASH_GET_APPLY_COUNT        ZEND_HASH_GET_APPLY_COUNT
+#define PHPC_HASH_INC_APPLY_COUNT        ZEND_HASH_INC_APPLY_COUNT
+#define PHPC_HASH_DEC_APPLY_COUNT        ZEND_HASH_DEC_APPLY_COUNT
+#define PHPC_HASH_APPLY_PROTECTION       ZEND_HASH_APPLY_PROTECTION
+#else
+#define PHPC_HASH_HAS_APPLY_COUNT        GC_IS_RECURSIVE
+#define PHPC_HASH_GET_APPLY_COUNT        GC_IS_RECURSIVE
+#define PHPC_HASH_INC_APPLY_COUNT        GC_PROTECT_RECURSION
+#define PHPC_HASH_DEC_APPLY_COUNT        GC_UNPROTECT_RECURSION
+#define PHPC_HASH_APPLY_PROTECTION(_tht) (!(GC_FLAGS(_tht) & GC_IMMUTABLE))
+#endif
 /* add ptr */
 #define PHPC_HASH_INDEX_ADD_PTR(_ht, _idx, _ptr, _ptr_size) \
 	zend_hash_index_add_ptr(_ht, _idx, _ptr)
@@ -969,6 +1081,11 @@ typedef zend_resource * phpc_res_value_t;
 #define PHPC_HASH_CSTR_ADD(_ht, _cstr_value, _pzv) \
 	zend_hash_str_add(_ht, _cstr_value, strlen(_cstr_value), _pzv)
 
+/* next insert */
+#define PHPC_HASH_NEXT_INDEX_INSERT_PTR(_ht, _ptr, _ptr_size) \
+	zend_hash_next_index_insert_ptr(_ht, _ptr)
+#define PHPC_HASH_NEXT_INDEX_INSERT zend_hash_next_index_insert
+
 /* update ptr */
 #define PHPC_HASH_INDEX_UPDATE_PTR(_ht, _idx, _ptr, _ptr_size) \
 	zend_hash_index_update_ptr(_ht, _idx, _ptr)
@@ -1052,6 +1169,12 @@ typedef zend_resource * phpc_res_value_t;
 #define PHPC_HASH_RETURN RETURN_ARR
 
 
+/* SYMTABLE */
+
+/* update*/
+#define PHPC_SYMTABLE_UPDATE zend_symtable_update
+
+
 /* ARRAY */
 
 #define PHPC_ARRAY_ADD_ASSOC_NULL_EX     add_assoc_null_ex
@@ -1106,29 +1229,44 @@ typedef zend_resource * phpc_res_value_t;
 /* ZVAL */
 typedef zval  phpc_val;
 
-#define PHPC_TYPE        Z_TYPE
-#define PHPC_TYPE_P      Z_TYPE_P
+#define PHPC_TYPE           Z_TYPE
+#define PHPC_TYPE_P         Z_TYPE_P
 
-#define PHPC_LVAL        Z_LVAL
-#define PHPC_LVAL_P      Z_LVAL_P
+#define PHPC_LVAL           Z_LVAL
+#define PHPC_LVAL_P         Z_LVAL_P
 
-#define PHPC_DVAL        Z_DVAL
-#define PHPC_DVAL_P      Z_DVAL_P
+#define PHPC_DVAL           Z_DVAL
+#define PHPC_DVAL_P         Z_DVAL_P
 
-#define PHPC_STRVAL      Z_STRVAL
-#define PHPC_STRVAL_P    Z_STRVAL_P
+#define PHPC_STRVAL         Z_STRVAL
+#define PHPC_STRVAL_P       Z_STRVAL_P
 
-#define PHPC_STRLEN      Z_STRLEN
-#define PHPC_STRLEN_P    Z_STRLEN_P
+#define PHPC_STRLEN         Z_STRLEN
+#define PHPC_STRLEN_P       Z_STRLEN_P
 
-#define PHPC_ARRVAL      Z_ARRVAL
-#define PHPC_ARRVAL_P    Z_ARRVAL_P
+#define PHPC_ARRVAL         Z_ARRVAL
+#define PHPC_ARRVAL_P       Z_ARRVAL_P
 
-#define PHPC_OBJCE       Z_OBJCE
-#define PHPC_OBJCE_P     Z_OBJCE_P
+#define PHPC_OBJ_HT         Z_OBJ_HT
+#define PHPC_OBJ_HT_P       Z_OBJ_HT_P
 
-#define PHPC_OBJPROP     Z_OBJPROP
-#define PHPC_OBJPROP_P   Z_OBJPROP_P
+#define PHPC_OBJ_HANDLER    Z_OBJ_HANDLER
+#define PHPC_OBJ_HANDLER_P  Z_OBJ_HANDLER_P
+
+#define PHPC_OBJ_HANDLE     Z_OBJ_HANDLE
+#define PHPC_OBJ_HANDLE_P   Z_OBJ_HANDLE_P
+
+#define PHPC_OBJCE          Z_OBJCE
+#define PHPC_OBJCE_P        Z_OBJCE_P
+
+#define PHPC_OBJPROP        Z_OBJPROP
+#define PHPC_OBJPROP_P      Z_OBJPROP_P
+
+#define PHPC_OBJDEBUG       Z_OBJDEBUG
+#define PHPC_OBJDEBUG_P     Z_OBJDEBUG_P
+
+#define PHPC_REFCOUNTED     Z_REFCOUNTED
+#define PHPC_REFCOUNTED_P   Z_REFCOUNTED_P
 
 #define PHPC_PZVAL_CAST_TO_PVAL(_pzv) _pzv
 
@@ -1147,9 +1285,17 @@ typedef zval  phpc_val;
 #define PHPC_VAL_COPY(_pv, _zv) ZVAL_COPY(&(_pv), _zv)
 #define PHPC_VAL_ASSIGN         PHPC_VAL_COPY
 
-#define PHPC_PVAL_DEREF         ZVAL_DEREF
-#define PHPC_PZVAL_DEREF        ZVAL_DEREF
+#define PHPC_PZVAL_MAKE(_pzv)    PHPC_NOOP
+#define PHPC_PZVAL_FREE(_pzv)    PHPC_NOOP
+#define PHPC_PZVAL_SET(_pv, _zv) _pv = _zv
+#define PHPC_PVAL_DEREF          ZVAL_DEREF
+#define PHPC_PZVAL_DEREF         ZVAL_DEREF
 
+#define PHPC_PZVAL_COPY_INIT     ZVAL_COPY_VALUE
+#define PHPC_PZVAL_COPY          ZVAL_COPY_VALUE
+
+#define PHPC_VAL_NEW_STR(_pv, _str) \
+	ZVAL_NEW_STR(&_pv, _str)
 #define PHPC_VAL_STR(_pv, _str) \
 	ZVAL_STR(&_pv, _str)
 #define PHPC_VAL_CSTR(_pv, _cstr) \
@@ -1157,9 +1303,18 @@ typedef zval  phpc_val;
 #define PHPC_VAL_CSTRL(_pv, _cstr, _cstr_len) \
 	ZVAL_STRINGL(&_pv, _cstr, _cstr_len)
 
-#define PHPC_PZVAL_STR   ZVAL_STR
-#define PHPC_PZVAL_CSTR  ZVAL_STRING
-#define PHPC_PZVAL_CSTRL ZVAL_STRINGL
+#define PHPC_PZVAL_NEW_STR ZVAL_NEW_STR
+#define PHPC_PZVAL_STR     ZVAL_STR
+#define PHPC_PZVAL_CSTR    ZVAL_STRING
+#define PHPC_PZVAL_CSTRL   ZVAL_STRINGL
+
+#define PHPC_ZVAL_COPY(_zv_dst, _zv_src) \
+	ZVAL_COPY_VALUE(&_zv_dst, &_zv_src)
+
+#define PHPC_ZVAL_NEW_STR  PHPC_VAL_NEW_STR
+#define PHPC_ZVAL_STR      PHPC_VAL_STR
+#define PHPC_ZVAL_CSTR     PHPC_VAL_CSTR
+#define PHPC_ZVAL_CSTRL    PHPC_VAL_CSTRL
 
 #define PHPC_ZVAL_IS_TRUE(_zv) \
 	(Z_TYPE(_zv) == IS_TRUE)
@@ -1215,8 +1370,13 @@ typedef const char phpc_stream_opener_char_t;
 /* COMMON (dependent definitions) */
 
 /* object structure */
+#define PHP_OBJ_GET_HANDLER_OBJ_FROM_ZOBJ(_name) \
+	PHPC_OBJ_FROM_ZOBJ(_name, _phpc_object)
 #define PHPC_OBJ_STRUCT_DECLARE_AND_FETCH_FROM_ZOBJ(_name, _ptr) \
 	PHPC_OBJ_STRUCT_DECLARE(_name, _ptr) = PHP_OBJ_GET_HANDLER_OBJ_FROM_ZOBJ(_name)
+#define PHPC_OBJ_FROM_SELF(_name) \
+	PHPC_OBJ_FROM_POBJ(_name, PHPC_SELF)
+
 
 /* this object */
 #define PHPC_THIS _phpc_this
@@ -1228,7 +1388,7 @@ typedef const char phpc_stream_opener_char_t;
 #define PHPC_THIS_DECLARE_AND_FETCH_FROM_ZVAL(_name, _zv) \
 	PHPC_THIS_DECLARE(_name) = PHPC_THIS_FETCH_FROM_ZVAL(_name, _zv)
 #define PHPC_THIS_FETCH_FROM_SELF(_name) \
-	PHPC_THIS = PHPC_OBJ_FROM_ZVAL(_name, PHPC_SELF)
+	PHPC_THIS = PHPC_OBJ_FROM_SELF(_name)
 #define PHPC_THIS_DECLARE_AND_FETCH_FROM_SELF(_name) \
 	PHPC_THIS_DECLARE(_name) = PHPC_OBJ_FROM_SELF(_name)
 #define PHPC_THIS_DECLARE_AND_FETCH(_name) \
@@ -1273,18 +1433,55 @@ typedef const char phpc_stream_opener_char_t;
 #define PHPC_GC_N _phpc_gc_n
 #define PHPC_OBJ_HANDLER_GET_GC(_name) \
 	PHPC_OBJ_DEFINE_HANDLER_FCE(HashTable *, _name, get_gc)\
-		(zval *PHPC_SELF, phpc_val **PHPC_GC_TABLE, int *PHPC_GC_N TSRMLS_DC)
+		(phpc_obj_t *PHPC_SELF, phpc_val **PHPC_GC_TABLE, int *PHPC_GC_N TSRMLS_DC)
 
 /* object handler get_debug_info */
 #define PHPC_DEBUG_INFO_IS_TEMP _phpc_debug_info_is_temp
 #define PHPC_OBJ_HANDLER_GET_DEBUG_INFO(_name) \
 	PHPC_OBJ_DEFINE_HANDLER_FCE(HashTable *, _name, get_debug_info)\
-		(zval *PHPC_SELF, int *PHPC_DEBUG_INFO_IS_TEMP TSRMLS_DC)
+		(phpc_obj_t *PHPC_SELF, int *PHPC_DEBUG_INFO_IS_TEMP TSRMLS_DC)
 
 /* object handler get_properties */
 #define PHPC_OBJ_HANDLER_GET_PROPERTIES(_name) \
 	PHPC_OBJ_DEFINE_HANDLER_FCE(HashTable *, _name, get_properties)\
-		(zval *PHPC_SELF TSRMLS_DC)
+		(phpc_obj_t *PHPC_SELF TSRMLS_DC)
+
+/* object handler setters */
+#define PHPC_CLASS_SET_HANDLER_CREATE(_class_entry, _name) \
+	_class_entry.create_object = PHPC_OBJ_GET_HANDLER_FCE(_name, create)
+#define PHPC_OBJ_SET_SPECIFIC_HANDLER_CLONE(_handlers, _name) \
+	(_handlers).clone_obj = PHPC_OBJ_GET_HANDLER_FCE(_name, clone)
+#define PHPC_OBJ_SET_HANDLER_CLONE(_name) \
+	PHPC_OBJ_SET_SPECIFIC_HANDLER_CLONE(PHPC_OBJ_GET_HANDLER_VAR_NAME(_name), _name)
+#define PHPC_OBJ_SET_HANDLER_COMPARE(_name) \
+	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).PHPC_OBJ_HANDLER_COMPARE_NAME \
+		= PHPC_OBJ_GET_HANDLER_FCE(_name, compare)
+#define PHPC_OBJ_HAS_HANDLER_GET_GC (PHP_VERSION_ID > 50399)
+#if PHPC_OBJ_HAS_HANDLER_GET_GC
+#define PHPC_OBJ_SET_HANDLER_GET_GC(_name) \
+	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).get_gc = PHPC_OBJ_GET_HANDLER_FCE(_name, get_gc)
+#else
+#define PHPC_OBJ_SET_HANDLER_GET_GC(_name) PHPC_NOOP
+#endif
+#define PHPC_OBJ_HAS_HANDLER_GET_DEBUG_INFO (PHP_VERSION_ID > 50299)
+#if PHPC_OBJ_HAS_HANDLER_GET_DEBUG_INFO
+#define PHPC_OBJ_SET_HANDLER_GET_DEBUG_INFO(_name) \
+	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).get_debug_info = PHPC_OBJ_GET_HANDLER_FCE(_name, get_debug_info)
+#else
+#define PHPC_OBJ_SET_HANDLER_GET_DEBUG_INFO(_name) PHPC_NOOP
+#endif
+/* there is such handler in 5.2 but we would have to re-implement zend_std_get_properties */
+#define PHPC_OBJ_HAS_HANDLER_GET_PROPERTIES (PHP_VERSION_ID > 50299)
+#if PHPC_OBJ_HAS_HANDLER_GET_PROPERTIES
+#define PHPC_OBJ_SET_HANDLER_GET_PROPERTIES(_name) \
+	PHPC_OBJ_GET_HANDLER_VAR_NAME(_name).get_properties = PHPC_OBJ_GET_HANDLER_FCE(_name, get_properties)
+#else
+#define PHPC_OBJ_SET_HANDLER_GET_PROPERTIES(_name) PHPC_NOOP
+#endif
+#define PHPC_OBJ_SET_HANDLER_OFFSET(_name) \
+	PHPC_OBJ_SET_SPECIFIC_HANDLER_OFFSET(PHPC_OBJ_GET_HANDLER_VAR_NAME(_name), _name)
+#define PHPC_OBJ_SET_HANDLER_FREE(_name) \
+	PHPC_OBJ_SET_SPECIFIC_HANDLER_FREE(PHPC_OBJ_GET_HANDLER_VAR_NAME(_name), _name)
 
 /* hash */
 #define PHPC_HASH_ALLOC                     ALLOC_HASHTABLE
@@ -1404,5 +1601,16 @@ typedef const char phpc_stream_opener_char_t;
 	php_stream_locate_url_wrapper(_path, (char **) _path_for_open, _options TSRMLS_CC)
 #endif
 
+#if (PHP_MAJOR_VERSION == 8 && PHP_MINOR_VERSION >= 1) || (PHP_MAJOR_VERSION > 8)
+#define PHPC_PHP_STAT(_filename, _filename_length, _type, _return_value) do { \
+		zend_string *_phpc_stat_str = zend_string_init(_filename, _filename_length, 0); \
+		php_stat(_phpc_stat_str, _type, _return_value); \
+		zend_string_release(_phpc_stat_str); \
+	} while (0)
+
+#else
+#define PHPC_PHP_STAT php_stat
+#endif
+
 #endif	/* PHPC_H */
 
